@@ -1,5 +1,3 @@
-// +build linux
-
 package main
 
 import (
@@ -96,6 +94,11 @@ using the runc checkpoint command.`,
 			Value: "",
 			Usage: "Specify an LSM profile to be used during restore in the form of TYPE:NAME.",
 		},
+		cli.StringFlag{
+			Name:  "lsm-mount-context",
+			Value: "",
+			Usage: "Specify an LSM mount context to be used during restore.",
+		},
 	},
 	Action: func(context *cli.Context) error {
 		if err := checkArgs(context, 1, exactArgs); err != nil {
@@ -106,15 +109,11 @@ using the runc checkpoint command.`,
 			logrus.Warn("runc checkpoint is untested with rootless containers")
 		}
 
-		spec, err := setupSpec(context)
-		if err != nil {
-			return err
-		}
 		options := criuOptions(context)
 		if err := setEmptyNsMask(context, options); err != nil {
 			return err
 		}
-		status, err := startContainer(context, spec, CT_ACT_RESTORE, options)
+		status, err := startContainer(context, CT_ACT_RESTORE, options)
 		if err != nil {
 			return err
 		}
@@ -145,5 +144,6 @@ func criuOptions(context *cli.Context) *libcontainer.CriuOpts {
 		LazyPages:               context.Bool("lazy-pages"),
 		StatusFd:                context.Int("status-fd"),
 		LsmProfile:              context.String("lsm-profile"),
+		LsmMountContext:         context.String("lsm-mount-context"),
 	}
 }
